import struct

def xor(key, buffer):
    """Returns an XOR encoded buffer"""
    if type(key) == int:
        xor_key = int(key)
    elif type(key) == NoneType:
        xor_key = 0
    else:
        xor_key = int(key, 16)

    if not (0 <= xor_key <= 255):
        raise ValueError("XOR key must be a value between 0 and 255")

    buf = ''

    for byte in bytearray(buffer):
        xor_byte = byte ^ xor_key
        buf += struct.pack('B', xor_byte)

    return buf

    


import copy, csv, ipaddress, os, re
# Python 2/3 switch
try:
    from itertools import izip_longest
except ImportError:
    from itertools import zip_longest as izip_longest

class Table(object):
    """Class used to pretty-print text in a tabular format"""
    def __init__(self, **kwargs):
        """
        Initializes a text table instance using provided arguments.
        
        Args:
            Header (str): The text to display above the table.  Default: None (no header)
            HeaderIndent (int): The number of spaces to indent the header.  Default: 0
            Columns (list): The list of columns to use in the table.
            Rows (list): The list of rows to use in the table.
            Width (int): The maximum character width of the table.  Default: 80
            Indent (int): The number of spaces to indent the table.  Default: 0
            CellPad (int): The number of spaces between each horizontal cell.  Default: 2
            Prefix (str): The text to display before the table.
            Postfix (str): The text to display after the table.
            SearchTerm (str): The regular expression used to filter rows.
            SortIndex (int): The column index used to sort the table.  -1 disables sorting.  Default: 0
            SortOrder (str): The direction in which to sort the table.  Default: 'forward'
        """
        if 'Header' in kwargs:
            self.header = kwargs['Header']
        else:
            self.header = None
            
        if 'HeaderIndent' in kwargs:
            self.headeri = kwargs['HeaderIndent']
        else:
            self.headeri = 0
        
        if 'Columns' in kwargs:
            self.columns = kwargs['Columns']
        else:
            self.columns = []
            
        self.rows = []
        
        if 'Width' in kwargs:
            self.width = kwargs['Width']
        else:
            self.width = 80
            
        if 'Indent' in kwargs:
            self.indent = kwargs['Indent']
        else:
            self.indent = 0
            
        if 'CellPad' in kwargs:
            self.cellpad = kwargs['CellPad']
        else:
            self.cellpad = 2
            
        if 'Prefix' in kwargs:
            self.prefix = kwargs['Prefix']
        else:
            self.prefix = ''
            
        if 'Postfix' in kwargs:
            self.postfix = kwargs['Postfix']
        else:
            self.postfix = ''
            
        self.colprops = []
        
        if 'SearchTerm' in kwargs:
            self.scterm = r'{}'.format(kwargs['SearchTerm'])
        else:
            self.scterm = None
            
        if 'SortIndex' in kwargs:
            self.sort_index = kwargs['SortIndex']
        else:
            self.sort_index = 0
            
        if 'SortOrder' in kwargs:
            self.sort_order = kwargs['SortOrder']
        else:
            self.sort_order = 'forward'
            
        for idx in range(0, len(self.columns)):
            self.colprops.append({})
            self.colprops[idx]['MaxWidth'] = len(self.columns[idx])
            
        if 'Rows' in kwargs:
            for row in kwargs['Rows']:
                self.add_row(row)
                
        if 'ColProps' in kwargs:
            for col in kwargs['ColProps']:
                try:
                    idx = self.columns.index(col)
                    self.colprops[idx].update(kwargs['ColProps'][col])
                except ValueError:
                    pass
                
            
    def add_row(self, fields = []):
        if len(fields) != len(self.columns):
            raise RuntimeError("Invalid number of columns!")
        
        for idx, field in enumerate(fields):
            field = str(field).strip()
            if self.colprops[idx]['MaxWidth'] < len(str(field)):
                old = self.colprops[idx]['MaxWidth']
                self.colprops[idx]['MaxWidth'] = len(str(field))
            
        self.rows.append(fields)
        
    def add_hr(self):
        self.rows.append('__hr__')
        
    @classmethod
    def ip_cmp(self, a, b):
        """Comparison operator used for IP addresses"""
        try:
            a = ipaddress.ip_address(unicode(str(a)))
            b = ipaddress.ip_address(unicode(str(b)))
            
            if a.version == 6 and b.version == 4:
                return 1
            elif a.version == 4 and b.version == 6:
                return -1
                
            if a < b:
                return -1
            elif a == b:
                return 0
            elif a > b:
                return 1
            else:
                return None
            
        except ipaddress.AddressValueError:
            return None
        except ValueError:
            return None

    def sort_rows(self, index = None, order = None):
        if index is None:
            index = self.sort_index
        if order is None:
            order = self.sort_order

        if index == -1:
            return

        if self.rows is None or len(self.rows) == 0:
            return

        def sort_block(a, b):
            try:
                a[index]
            except IndexError:
                cmp = -1
            try:
                b[index]
            except IndexError:
                cmp = 1
            if a[index] is None:
                cmp = -1
            elif b[index] is None:
                cmp = 1
            elif (re.search(r'^[0-9]+$', a[index]) is not None) and (re.search(r'^[0-9]+$', b[index]) is not None):
                if int(a[index]) < int(b[index]):
                    cmp = -1
                elif int(a[index]) == int(b[index]):
                    cmp = 0
                elif int(a[index]) > int(b[index]):
                    cmp = 1
                else:
                    cmp = None
            elif (self.ip_cmp(a[index], b[index]) is not None):
                cmp = self.ip_cmp(a[index], b[index])
            else:
                if (a[index]) < (b[index]):
                    cmp = -1
                elif (a[index]) == (b[index]):
                    cmp = 0
                elif (a[index]) > (b[index]):
                    cmp = 1
                else:
                    cmp = None
            if cmp is None:
                cmp = 0
            if order == 'forward':
                return cmp
            else:
                return -cmp

        self.rows = sorted(self.rows, cmp = sort_block)
            
    def valid_ip(self, value):
        try:
            ipaddress.ip_address(unicode(value))
            return True
        except ipaddress.AddressValueError:
            return False
        except ValueError:
            return False
            
    @classmethod
    def from_csv(self, csvobj):
        """Returns a Table instance using CSV data (either from a string or a file)"""
        if type(csvobj) is str:
            # was it a filename or a raw CSV string?
            if os.path.isfile(csvobj) == True:
                csvfile = file(csvobj, 'rb')
                csvobj = list(csv.reader(csvfile))
            else:
                csvobj = list(csv.reader(csvobj.splitlines()))
                
        if csvobj[0] == ["Keys", "Values"]:
            csvobj.pop(0)
            cols = []
            rows = []
            for row in csvobj:
                cols.append(row.pop(0))
                rows.append(row)
            
            tbl = self.__class__(Columns=cols)
            groups = list(izip_longest(*[iter(rows)] * len(cols)))
            for r in groups:
                tbl.add_row(list(r))
                
        else:
            tbl = self.__class__(Columns=csvobj.pop(0))
            while len(csvobj) > 0:
                tbl.add_row(csvobj.pop(0))
                
        return tbl

    def to_csv(self):
        """Converts the table into CSV format"""
        def csv_block(x):
            x = str(x)
            x = re.sub(r'[\r\n]', ' ', x)
            x = re.sub(r'\s+', ' ', x)
            x = re.sub(r'"', '""', x)
            return x

        string = ''
        string += (",".join(self.columns) + "\n")
        for row in self.rows:
            if self.is_hr(row) == True or self.row_visible(row) == False:
                continue
            string += (",".join(map(lambda x: "\"{}\"".format(x), map(lambda x: csv_block(x), row))) + "\n")

        return string   
        
    def __str__(self):
        """Returns string representation of the table"""
        retval = copy.copy(self.prefix)
        retval = retval + self.header_to_str()
        retval = retval + self.columns_to_str()
        retval = retval + self.hr_to_str()
        
        self.sort_rows()
        for row in self.rows:
            if self.is_hr(row) == True:
                retval = retval + self.hr_to_str()
            else:
                if self.row_visible(row):
                    retval = retval + self.row_to_str(row)
                    
        retval = retval + self.postfix
        
        return retval
        
    def header_to_str(self):
        if self.header is not None:
            pad = " " * self.headeri
            return pad + self.header + "\n" + pad + ("=" * len(self.header)) + "\n\n"
        
        return ''
        
    def columns_to_str(self):
        nameline = ' ' * self.indent
        barline = copy.copy(nameline)
        last_col = None
        last_idx = None
        for idx, col in enumerate(self.columns):
            if last_col is not None:
                padding = self.pad(' ', last_col, last_idx)
                nameline = nameline + padding
                remainder = len(padding) - self.cellpad
                if remainder < 0:
                    remainder = 0
                barline = barline + (' ' * (self.cellpad + remainder))
                
            nameline = nameline + col
            barline = barline + ('-' * len(col))
            
            last_col = col
            last_idx = idx
            
        return "{}\n{}".format(nameline, barline)
        
    def hr_to_str(self):
        return "\n"
        
    def row_to_str(self, row):
        line = ' ' * self.indent
        last_cell = None
        last_idx = None
        for idx, cell in enumerate(row):
            if idx != 0:
                line = line + self.pad(' ', str(last_cell), last_idx)
            
            if 'MaxChar' in self.colprops[idx]:
                last_cell = str(cell)[0:(int(self.colprops[idx]['MaxChar']) + 1)]
                line = line + last_cell
            else:
                line = line + str(cell)
                last_cell = cell
                
            last_idx = idx
            
        return line + "\n"
        
    def pad(self, chr, buf, colidx, use_cell_pad = True):
        if 'MaxChar' in self.colprops[colidx]:
            max = self.colprops[colidx]['MaxChar']
        elif 'MaxWidth' in self.colprops[colidx]:
            max = self.colprops[colidx]['MaxWidth']
            
        if int(max) > int(self.colprops[colidx]['MaxWidth']):
            max = self.colprops[colidx]['MaxWidth']
            
        strcpy = str(copy.copy(buf))
        remainder = max - len(strcpy)
        if remainder < 0:
            remainder = 0
            
        val = chr * remainder
        
        if use_cell_pad == True:
            val = val + (' ' * self.cellpad)
            
        return val
        
    def row_visible(self, row):
        if self.scterm is None:
            return True

        return (re.search(self.scterm, self.row_to_str(row)) is not None)
            
    def is_hr(self, row):
        return isinstance(row, str) and (row == '__hr__')
                
    def __getitem__(self, col_names):
        if type(col_names) == str:
            col_names = [col_names]

        tbl = self.__class__(Indent=self.indent,Header=self.header,Columns=col_names)

        indices = []

        for col_name in col_names:
            try:
                index = self.columns.index(col_name)
            except ValueError:
                raise ValueError("Invalid column name '{}'".format(col_name))
            indices.append(index)

        for old_row in self.rows:
            new_row = []
            map(lambda i: new_row.append(old_row[i]), indices)
            tbl.add_row(new_row)

        return tbl

        


from ctypes import *
from struct import *
import os, sys, re

class WinColorConsole(object):
    """Class that provides methods to output color on Windows-based consoles"""
    STD_OUTPUT_HANDLE = -11
    COLORS = [0, 4, 2, 6, 1, 5, 3, 7]
    
    def __init__(self, stream):
        self.stream = stream
        # windows specific stuff from kernel32 API
        self.GetStdHandle = windll.kernel32.GetStdHandle
        self.GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
        self.SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
        self.hConsoleHandle = self.GetStdHandle(self.STD_OUTPUT_HANDLE)

    def write(self, msg):
        rest = msg
        colorlist = re.findall(r'\x1B\[[^A-Za-z]*[A-Za-z]', msg)
        pos = 0
        for i in colorlist:
            color = i[i.find('[') + 1 : i.find('m')]
            pos = rest.find(i)
            if pos == 0:
                # the color code is the first part of the string
                # change color and slice the bytes, next color
                self.setcolor(int(color))
                rest = rest.replace(i, '')
                continue
            elif pos > 0:
                # otherwise write the color text
                # then change color again
                self.stream.write(rest[:pos])
                self.setcolor(int(color))
                rest = rest.replace(i , '')
                continue
                
        self.stream.write(rest[pos:])
        
    def flush(self):
        self.stream.flush()

    def setcolor(self, color):
        csbi = chr(0) * 22
        self.GetConsoleScreenBufferInfo(self.hConsoleHandle, csbi)
        wAttr = unpack('H', csbi[8:10])[0]
        
        if int(color) == 0: # reset
            wAttr = 0x07
        elif int(color) == 1: # bold
            wAttr |= 0x08
        elif int(color) == 2: # unbold
            wAttr &= -0x09
        elif int(color) == 7: # reverse
            wAttr = ((wAttr & 0x0f) << 4) | ((wAttr & 0xf0) >> 4)
        elif int(color) == 8: # conceal
            wAttr &= -16
        elif int(color) in range(30, 38): # foreground colors
            wAttr = ((wAttr & (-8)) | (self.COLORS[color - 30]))
        elif int(color) in range(40, 48): # background colors
            wAttr = ((wAttr & (-113)) | (self.COLORS[color - 40] << 4))
        
        self.SetConsoleTextAttribute(self.hConsoleHandle, wAttr)	



import random, struct

UpperAlpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
LowerAlpha = "abcdefghijklmnopqrstuvwxyz"
Numerals = "0123456789"
Base32 = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
Base64 = UpperAlpha + LowerAlpha + Numerals + '+/'
Base64Url = UpperAlpha + LowerAlpha + Numerals + '-_'
Alpha = UpperAlpha + LowerAlpha
AlphaNumeric = Alpha + Numerals
DefaultWrap = 60

try:
    HighAscii = struct.pack("128B", *range(0x80, 0xff + 1))
    LowAscii = struct.pack("32B", *range(0x00, 0x1f + 1))
    AllChars = struct.pack("256B", *range(0x00, 0xff + 1))
    Punctuation = struct.pack("31B", *(range(0x21, 0x2f + 1) + range(0x3a, 0x3F + 1) + range(0x5b, 0x60 + 1) + range(0x7b, 0x7e + 1)))
except Exception:
    HighAscii = struct.pack("128B", *list(range(0x80, 0xff + 1)))
    LowAscii = struct.pack("32B", *list(range(0x00, 0x1f + 1)))
    AllChars = struct.pack("256B", *list(range(0x00, 0xff + 1)))
    Punctuation = struct.pack("31B", *(list(range(0x21, 0x2f + 1)) + list(range(0x3a, 0x3F + 1)) + list(range(0x5b, 0x60 + 1)) + list(range(0x7b, 0x7e + 1))))

TLDs = ['com', 'net', 'org', 'gov', 'biz', 'edu']

States = [
    "AK", "AL", "AR", "AZ", "CA", "CO", "CT", "DE", "FL", "GA", "HI",
    "IA", "ID", "IL", "IN", "KS", "KY", "LA", "MA", "MD", "ME", "MI", "MN",
    "MO", "MS", "MT", "NC", "ND", "NE", "NH", "NJ", "NM", "NV", "NY", "OH",
    "OK", "OR", "PA", "RI", "SC", "SD", "TN", "TX", "UT", "VA", "VT", "WA",
    "WI", "WV", "WY"
    ]
    
Surnames = [
      "adams", "alexander", "allen", "anderson", "bailey", "baker", "barnes",
      "bell", "bennett", "brooks", "brown", "bryant", "butler", "campbell",
      "carter", "clark", "coleman", "collins", "cook", "cooper", "cox",
      "davis", "diaz", "edwards", "evans", "flores", "foster", "garcia",
      "gonzales", "gonzalez", "gray", "green", "griffin", "hall", "harris",
      "hayes", "henderson", "hernandez", "hill", "howard", "hughes", "jackson",
      "james", "jenkins", "johnson", "jones", "kelly", "king", "lee", "lewis",
      "long", "lopez", "martin", "martinez", "miller", "mitchell", "moore",
      "morgan", "morris", "murphy", "nelson", "parker", "patterson", "perez",
      "perry", "peterson", "phillips", "powell", "price", "ramirez", "reed",
      "richardson", "rivera", "roberts", "robinson", "rodriguez", "rogers",
      "ross", "russell", "sanchez", "sanders", "scott", "simmons", "smith",
      "stewart", "taylor", "thomas", "thompson", "torres", "turner", "walker",
      "ward", "washington", "watson", "white", "williams", "wilson", "wood",
      "wright", "young"
    ]
    
Names_Male = [
      "aaron", "adam", "alan", "albert", "andrew", "anthony", "antonio",
      "arthur", "benjamin", "billy", "bobby", "brandon", "brian", "bruce",
      "carl", "carlos", "charles", "chris", "christopher", "clarence", "craig",
      "daniel", "david", "dennis", "donald", "douglas", "earl", "edward",
      "eric", "ernest", "eugene", "frank", "fred", "gary", "george", "gerald",
      "gregory", "harold", "harry", "henry", "howard", "jack", "james", "jason",
      "jeffrey", "jeremy", "jerry", "jesse", "jimmy", "joe", "john", "johnny",
      "jonathan", "jose", "joseph", "joshua", "juan", "justin", "keith",
      "kenneth", "kevin", "larry", "lawrence", "louis", "mark", "martin",
      "matthew", "michael", "nicholas", "patrick", "paul", "peter", "philip",
      "phillip", "ralph", "randy", "raymond", "richard", "robert", "roger",
      "ronald", "roy", "russell", "ryan", "samuel", "scott", "sean", "shawn",
      "stephen", "steve", "steven", "terry", "thomas", "timothy", "todd",
      "victor", "walter", "wayne", "william", "willie"
    ]
    
Names_Female = [
      "alice", "amanda", "amy", "andrea", "angela", "ann", "anna", "anne",
      "annie", "ashley", "barbara", "betty", "beverly", "bonnie", "brenda",
      "carol", "carolyn", "catherine", "cheryl", "christina", "christine",
      "cynthia", "deborah", "debra", "denise", "diana", "diane", "donna",
      "doris", "dorothy", "elizabeth", "emily", "evelyn", "frances", "gloria",
      "heather", "helen", "irene", "jacqueline", "jane", "janet", "janice",
      "jean", "jennifer", "jessica", "joan", "joyce", "judith", "judy", "julia",
      "julie", "karen", "katherine", "kathleen", "kathryn", "kathy", "kelly",
      "kimberly", "laura", "lillian", "linda", "lisa", "lois", "lori", "louise",
      "margaret", "maria", "marie", "marilyn", "martha", "mary", "melissa",
      "michelle", "mildred", "nancy", "nicole", "norma", "pamela", "patricia",
      "paula", "phyllis", "rachel", "rebecca", "robin", "rose", "ruby", "ruth",
      "sandra", "sara", "sarah", "sharon", "shirley", "stephanie", "susan",
      "tammy", "teresa", "theresa", "tina", "virginia", "wanda"
    ]
    
def rand_base(length, bad, *foo):
    """Random text generator"""
    list_one = list((struct.unpack("{}B".format(len("".join(foo))), "".join(foo))))
    list_two = list((struct.unpack("{}B".format(len(str(bad))), str(bad))))
    
    cset = [x for x in list_one if x not in list_two]
    cset = list(set(cset))
    
    outp = []
    
    if len(cset) == 0:
        return ""
        
    for i in range(0, length):
        outp.append(random.choice(cset))
        
    return struct.pack("{}B".format(length), *outp)
    
def rand_text(length, bad = '', chars = AllChars):
    """Returns random bytes of data"""
    foo = chars.split()
    return rand_base(length, bad, *foo)

def rand_char(bad = '', chars = AllChars):
    """Returns a random character"""
    return rand_text(1, bad, chars)
    
def rand_text_alpha(length, bad = ''):
    """Returns random bytes of alphabetical data"""
    foo = []
    foo = foo + list(UpperAlpha)
    foo = foo + list(LowerAlpha)
    return rand_base(length, bad, *foo)
    
def rand_text_alpha_lower(length, bad = ''):
    """Returns random bytes of lowercase alphabetical data"""
    foo = []
    foo = foo + list(LowerAlpha)
    return rand_base(length, bad, *foo)
    
def rand_text_alpha_upper(length, bad = ''):
    """Returns random bytes of uppercase alphabetical data"""
    foo = []
    foo = foo + list(UpperAlpha)
    return rand_base(length, bad, *foo)
    
def rand_text_alphanumeric(length, bad = ''):
    """Returns random bytes of alphanumeric data"""
    foo = []
    foo = foo + list(UpperAlpha)
    foo = foo + list(LowerAlpha)
    foo = foo + list(Numerals)
    return rand_base(length, bad, *foo)
    
def rand_text_hex(length, bad = ''):
    """Returns random bytes of alphanumeric hex data"""
    foo = []
    foo = foo + list(Numerals)
    foo = foo + list(LowerAlpha[0:6])
    return rand_base(length, bad, *foo)
    
def rand_text_numeric(length, bad = ''):
    """Returns random bytes of numeric data"""
    foo = list(Numerals)
    return rand_base(length, bad, *foo)

def rand_hostname():
    """Returns a random domain name"""
    host = []
    for i in range(0, random.randrange(3) + 1):
        host.append(rand_text_alphanumeric(random.randrange(10) + 1))
        
    host.append(random.choice(TLDs))
    return '.'.join(host).lower()
    
def rand_state():
    """Returns a random US state name"""
    return random.choice(States)
    
def rand_surname():
    """Returns a random last name"""
    return random.choice(Surnames)
    
def rand_name():
    """Returns a random male or female name"""
    if random.randrange(10) % 2 == 0:
        return random.choice(Names_Male)
    else:
        return random.choice(Names_Female)
        
def rand_name_male():
    """Returns a random male name"""
    return random.choice(Names_Male)
    
def rand_name_female():
    """Returns a random female name"""
    return random.choice(Names_Female)
    
def rand_mail_address():
    """Returns a random email address"""
    return str(rand_name() + '.' + rand_surname() + '@' + rand_hostname())
    
def rand_guid():
    """Returns a random GUID"""
    return "{" + "{}".format("-".join(map(lambda a: rand_text_hex(a), [8, 4, 4, 4, 12]))) + "}"    

def to_rand_case(string):
    """Randomizes capitalization of a string"""
    buf = []
    for i in range(0, len(string)):
        if random.randrange(2) == 0:
            buf.append(string[i].upper())
        else:
            buf.append(string[i].lower())
    return "".join(buf)
    



try:
    import hex
except ImportError:
    from . import hex
import struct
try:
    import wordwrap
except ImportError:
    from . import wordwrap

DefaultWrap = 60

def to_python(string, wrap = DefaultWrap, name = "buf"):
    """Converts a raw string into a Python buffer"""
    return hex.hexify(string, wrap, "{} += \"".format(name), '"', "{} = \"\"\n".format(name), '"')

def to_python_comment(string, wrap = DefaultWrap):
    """Creates a Python comment"""
    return wordwrap.wordwrap(string, 0, wrap, '', '# ')

def to_c(string, wrap = DefaultWrap, name = "buf"):
    """Converts a raw string into a C-style byte array"""
    return hex.hexify(string, wrap, '"', '"', "unsigned char {}[] = \n".format(name), '";')

def to_c_comment(string, wrap = DefaultWrap):
    """Creates a C comment"""
    return "/*\n" + wordwrap.wordwrap(string, 0, wrap, '', ' * ') + " */\n"

def to_vbscript(string, name = "buf"):
    """Converts a raw string into a VBScript byte array"""
    if string is None or len(string) == 0:
        return "{}".format(name)

    code = struct.unpack("{}B".format(len(string)), string)
    buff = "{}=Chr({})".format(name, code[0])
    for byte in range(1, len(code)):
        if (byte % 100 == 0):
            buff += "\r\n{}={}".format(name, name)

        buff += "&Chr({})".format(code[byte])

    return buff

def to_vba(string, name = "buf"):
    """Converts a raw string into a VBA byte array"""
    if string is None or len(string) == 0:
        return "{} = Array()".format(name)

    code = struct.unpack("{}B".format(len(string)), string)
    buff = "{} = Array(".format(name)
    maxbytes = 80

    for idx in range(0, len(code)):
        buff += str(code[idx])
        if idx < len(code) - 1:
            buff += ","
        if (idx > 1 and (idx % maxbytes) == 0):
            buff += " _\r\n"

    buff += ")\r\n"

    return buff




import re

DefaultWrap = 60

def wordwrap(string, indent = 0, col = DefaultWrap, append = '', prepend = ''):
    """Wraps text at a given column using the provided indentation value"""
    def wrap_block(w):
        retval = (" " * indent) + prepend + w + append + chr(5)
        retval = re.sub(r'\n\005', "\n", retval)
        retval = re.sub(r'\005', "\n", retval)
        return retval

    return re.sub(r'.{1,' + "{}".format(col - indent) + '}(?:\s|\Z)', lambda w: wrap_block(w.group()), string)


import base64, re

def encode_base64(string, delim = ''):
    return re.sub(r'\s+', delim, (base64.b64encode(str(string)) + "\n"))
    
def decode_base64(string):
    return base64.b64decode(str(string))
    
def encode_base64url(string, delim = ''):
    return re.sub('=', '', encode_base64(string, delim).replace('+/', '-_'))
    
def decode_base64url(string):
    return decode_base64(re.sub(r'[^a-zA-Z0-9_\-]', '', string).replace('-_', '+/'))


import binascii, re, struct

DefaultWrap = 60

def to_hex(string, prefix = "\\x", count = 1):
    """Returns the hex version of the provided string"""
    """The count parameter defines the bytes per character (default: 1 byte)"""
    if ((len(string) % count) > 0):
        raise RuntimeError("Unable to chunk into {} byte chunks".format(count))
    
    return re.sub(r".{" + str((count * 2)) + "}", lambda s: prefix + s.group(), binascii.hexlify(string))

def to_hex_ascii(string, prefix = "\\x", count = 1, suffix = ""):
    """Returns the string with nonprintable ASCII characters converted to hex"""
    if ((len(string) % count) > 0):
        raise RuntimeError("Unable to chunk into {} byte chunks".format(count))

    def hex_block(s):
        if int(s, 16) in range(0x20, 0x7e + 1):
            return chr(int(s, 16))
        else:
            return prefix + s + str(suffix)

    return re.sub(r".{" + str((count * 2)) + "}", lambda s: hex_block(s.group()), binascii.hexlify(string))

def hex_to_raw(string):
    """Returns a raw string from a hex encoded string"""
    return binascii.unhexlify(string)

def hexify(string, col = DefaultWrap, line_start = '', line_end = '', buf_start = '', buf_end = ''):
    """Returns the hex version of the provided string (similar to to_hex(), but with word wrap support)"""
    output = buf_start
    cur = 0
    count = 0
    new_line = True

    for byte in bytearray(string):
        count += 1
        append = ''

        if (new_line == True):
            append += line_start
            new_line = False

        append += "\\x{:02x}".format(byte)
        cur += len(append)

        if ((cur + len(line_end) >= col) or (cur + len(buf_end) >= col)):
            new_line = True
            cur = 0

            if (count == len(string)):
                append += buf_end + "\n"
            else:
                append += line_end + "\n"

        output += append

    if (new_line == False):
        output += buf_end + "\n"

    return output

def hexdump(string, width = 16, base = 0):
    """Returns a hex dump of a string"""
    buf = ''
    idx = 0
    cnt = 0
    snl = False
    lst = 0
    lft_col_len = len("{:x}".format(int(base) + len(string)))
    if lft_col_len < 8:
        lft_col_len = 8

    while (idx < len(string)):
        chunk = string[idx:idx+width]
        addr = ("{:0" + "{}".format(lft_col_len) + "x}  ").format(int(base) + idx)
        line = " ".join(re.findall(r'..', binascii.hexlify(chunk)))
        buf += addr + line

        if lst == 0:
            lst = len(line)
            buf += (" " * 4)
        else:
            buf += (" " * abs(((lst - len(line)) + 4)))

        buf += "|"
        for c in struct.unpack("{}B".format(len(chunk)), chunk):
            if (c > 0x1f and c < 0x7f):
                buf += chr(c)
            else:
                buf += "."

        buf += "|\n"

        idx += width

    return buf + "\n"
        

def dehex(string):
    """Replaces hex-encoded characters with their ASCII equivalents"""
    if string is None:
        return string

    regex = r'\x5cx[0-9a-f]{2}'
    return re.sub(regex, lambda x: chr(int(x.group()[2:4], 16)), string, flags = re.M | re.I)
    